/*
 * Copyright (c) 2015 N. Billard.
 */
package com.massa.mrules.extensions.jdk8.addon;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAccessor;

import com.massa.mrules.addon.DefaultVariablesManager;
import com.massa.mrules.addon.DefaultVariablesManager.AbstractVariableAccessor;
import com.massa.mrules.addon.IAddonsManager;
import com.massa.mrules.addon.MAddonsFinder;
import com.massa.mrules.context.execute.IExecutionContext;
import com.massa.mrules.exception.MAccessorException;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MRuleInternalException;
import com.massa.util.ConfigDiscovery;
import com.massa.util.MBeanUtils;
import com.massa.util.UtilsException;

/**
 * Declares new system variables allowing to handle dates and times.
 * <ul>
 *   <li><b>DATE</b>: Current date.</li>
 *   <li><b>DATETIME</b>: Current date + time.</li>
 *   <li><b>TIME</b>: Current time.</li>
 *   <li><b>NOW</b>: Alias for datetime.</li>
 * </ul>
 * @see com.massa.mrules.addon.DefaultVariablesManager
 */
public class Jdk8VariablesManager implements IAddonsManager {

	/**
	 * Manually registers variables accessors using {@link ConfigDiscovery}.
	 * Variable Accessors implementation ID <b>must</b> start with the "#" character.
	 */
	@Override
	public void registerAddons() throws MConfigurationException, MRuleInternalException {
		final Jdk8VariablesManager.Jdk8Date date = new Jdk8VariablesManager.Jdk8Date();
		final Jdk8VariablesManager.Jdk8DateTime dateTime = new Jdk8VariablesManager.Jdk8DateTime();
		final Jdk8VariablesManager.Jdk8Time time = new Jdk8VariablesManager.Jdk8Time();
		final Jdk8VariablesManager.Jdk8Now now = new Jdk8VariablesManager.Jdk8Now();

		MAddonsFinder.unregisterAddon(date.getImplementationId());
		MAddonsFinder.unregisterAddon(dateTime.getImplementationId());
		MAddonsFinder.unregisterAddon(time.getImplementationId());
		MAddonsFinder.unregisterAddon(now.getImplementationId());

		MAddonsFinder.registerAddon(date);
		MAddonsFinder.registerAddon(dateTime);
		MAddonsFinder.registerAddon(time);
		MAddonsFinder.registerAddon(now);
	}

	private static TemporalAccessor getTemporal(final IExecutionContext ctxt, final AbstractVariableAccessor v) throws MAccessorException {
		Class<?> tclass = null;
		if (v.getType(ctxt) != null && TemporalAccessor.class.isAssignableFrom(v.getType(ctxt))) {
			tclass = v.getType(ctxt);
		}
		if (v.getGenericCollectionType(ctxt) != null && TemporalAccessor.class.isAssignableFrom(v.getGenericCollectionType(ctxt))) {
			tclass = v.getGenericCollectionType(ctxt);
		}
		if (tclass != null) {
			try {
				return (TemporalAccessor) MBeanUtils.staticCall(tclass, "now");
			} catch (final UtilsException e) {
				try {
					return (TemporalAccessor) MBeanUtils.staticCall(tclass, "from", new Class[]{TemporalAccessor.class}, LocalDateTime.now());
				} catch (final UtilsException se) {
					//Silence... Will try conversion
				}
			}
		}
		return null;
	}

	/**
	 * Default "#DATE" system variable accessor.
	 */
	public static class Jdk8Date extends DefaultVariablesManager.Date {

		private static final long serialVersionUID = 1L;

		@Override
		public Object getVarValue(final IExecutionContext ctxt) throws MAccessorException {
			final TemporalAccessor t = Jdk8VariablesManager.getTemporal(ctxt, this);
			return t == null ? super.getVarValue(ctxt) : t;
		}
	}

	/**
	 * Default "#DATETIME" system variable accessor.
	 */
	public static class Jdk8DateTime extends DefaultVariablesManager.DateTime {

		private static final long serialVersionUID = 1L;

		@Override
		public Object getVarValue(final IExecutionContext ctxt) throws MAccessorException {
			final TemporalAccessor t = Jdk8VariablesManager.getTemporal(ctxt, this);
			return t == null ? super.getVarValue(ctxt) : t;
		}
	}

	/**
	 * Default "#TIME" system variable accessor.
	 */
	public static class Jdk8Time extends DefaultVariablesManager.Time {

		private static final long serialVersionUID = 1L;

		@Override
		public Object getVarValue(final IExecutionContext ctxt) throws MAccessorException {
			final TemporalAccessor t = Jdk8VariablesManager.getTemporal(ctxt, this);
			return t == null ? super.getVarValue(ctxt) : t;
		}
	}

	/**
	 * Default "#NOW" system variable accessor.
	 */
	public static class Jdk8Now extends Jdk8DateTime {

		private static final long serialVersionUID = 1L;

		@Override
		public String getImplementationId() {
			return DefaultVariablesManager.GLOBAL_VAR_PREFIX + "NOW";
		}
	}
}
