package com.massa.util.extensions.jdk8.convert;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.time.DateTimeException;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.LinkedHashSet;

import javax.xml.datatype.XMLGregorianCalendar;

import com.massa.util.UtilsException;
import com.massa.util.convert.ConvertUtils;
import com.massa.util.convert.DateConverter;

/**
 * Converter handling new JDK 8 java.time package.
 */
public class Jdk8DateTimeConverter extends DateConverter {

	@Override
	protected String internalConvertToString(final Object o) throws UtilsException {
		if (o instanceof TemporalAccessor) {
			DateTimeFormatter f = null;
			if (this.getPreferredPattern() != null) {
				if (this.getLocale() != null) {
					f = DateTimeFormatter.ofPattern(this.getPreferredPattern(), this.getLocale());
				} else {
					f = DateTimeFormatter.ofPattern(this.getPreferredPattern());
				}
			}
			if (f != null) {
				return f.format((TemporalAccessor) o) ;
			} else {
				return o.toString();
			}
		}
		return super.internalConvertToString(o);
	}

	@SuppressWarnings({"unchecked", "rawtypes"})
	@Override
	protected <T> T internalConvertToType(final Class<T> type, final Object o) throws UtilsException {
		if (TemporalAccessor.class.isAssignableFrom(type)) {
			TemporalAccessor dt = null;

			if (o instanceof Date) {
				dt = this.convertDateToTemporal(type, (Date) o);
			} else if (o instanceof Calendar) {
				dt = this.convertDateToTemporal(type, ((Calendar) o).getTime());
			} else if (o instanceof XMLGregorianCalendar) {
				dt = this.convertDateToTemporal(type, ((XMLGregorianCalendar) o).toGregorianCalendar().getTime());
			} else if (o instanceof Long) {
				dt = this.convertDateToTemporal(type, new Date((Long) o));
			} else if (o instanceof TemporalAccessor) {
				dt = this.convertTemporalToTemporal(type, (TemporalAccessor) o);
			} else {
				final String stringValue = ConvertUtils.getDefaultInstance().lookup(String.class).convertToString(o);

				try {
					if (type.isEnum()) {
						try {
							//Mandatory intermediate var, as otherwise compile fails with some compiler versions.
							final Enum valueOf = Enum.valueOf((Class) type, stringValue);
							return type.cast(valueOf);
						} catch (final Exception e) {
							//Silence
						}
					}
					final Method m = type.getDeclaredMethod("from", TemporalAccessor.class);
					if (this.getPreferredPattern() != null) {
						dt = this.tryPattern(type, stringValue, m, this.getPreferredPattern());
					}
					if (dt == null) {
						dt = this.tryPatterns(type, stringValue, m, DateConverter.getGlobalpatterns());
					}
					if (dt == null) {
						dt = this.tryPatterns(type, stringValue, m, this.getSpecificpatterns());
					}
				} catch (NoSuchMethodException | RuntimeException e1) {
					//Silence
				}
			}

			if (dt == null) {
				throw this.convertionException(o, type);
			}
			return type.cast(dt);
		}
		if (o instanceof TemporalAccessor) {
			final Calendar c = this.convertTemporalToCalendar((TemporalAccessor) o);

			if (type.isAssignableFrom(Calendar.class)) {
				return type.cast(c);
			}
			return super.internalConvertToType(type, c);
		}
		return super.internalConvertToType(type, o);
	}

	private TemporalAccessor tryPattern(final Class<?> type, final String stringValue, final Method m, final String pattern) {
		DateTimeFormatter f;
		if (this.getLocale() != null) {
			f = DateTimeFormatter.ofPattern(pattern, this.getLocale());
		} else {
			f = DateTimeFormatter.ofPattern(pattern);
		}
		try {
			return f.parse(stringValue, x -> this.convertTemporalToTemporal(type, x, m));
		} catch (final DateTimeParseException e) {
			//Silence
		}
		return null;
	}

	private TemporalAccessor tryPatterns(final Class<?> type, final String stringValue, final Method m, final LinkedHashSet<String> patterns) {
		for (final String pattern: patterns) {
			final TemporalAccessor dt = this.tryPattern(type, stringValue, m, pattern);
			if (dt != null) {
				return dt;
			}
		}
		return null;
	}

	protected Calendar convertTemporalToCalendar(final TemporalAccessor ta) {
		int year = 1970;
		int month = 1;
		int day = 1;
		int hour = 0;
		int min = 0;
		int sec = 0;
		int ms = 0;
		if (ta.isSupported(ChronoField.YEAR)) {
			year = ta.get(ChronoField.YEAR);
		}
		if (ta.isSupported(ChronoField.MONTH_OF_YEAR)) {
			month = ta.get(ChronoField.MONTH_OF_YEAR);
		}
		if (ta.isSupported(ChronoField.DAY_OF_MONTH)) {
			day = ta.get(ChronoField.DAY_OF_MONTH);
		}
		if (ta.isSupported(ChronoField.HOUR_OF_DAY)) {
			hour = ta.get(ChronoField.HOUR_OF_DAY);
		}
		if (ta.isSupported(ChronoField.MINUTE_OF_HOUR)) {
			min = ta.get(ChronoField.MINUTE_OF_HOUR);
		}
		if (ta.isSupported(ChronoField.SECOND_OF_MINUTE)) {
			sec = ta.get(ChronoField.SECOND_OF_MINUTE);
		}
		if (ta.isSupported(ChronoField.MILLI_OF_SECOND)) {
			ms = ta.get(ChronoField.MILLI_OF_SECOND);
		}

		final Calendar c = Calendar.getInstance();
		c.set(Calendar.YEAR, year);
		c.set(Calendar.MONTH, month - 1);
		c.set(Calendar.DATE, day);
		c.set(Calendar.HOUR_OF_DAY, hour);
		c.set(Calendar.MINUTE, min);
		c.set(Calendar.SECOND, sec);
		c.set(Calendar.MILLISECOND, ms);
		return c;
	}

	protected <T> TemporalAccessor convertDateToTemporal(final Class<T> type, final Date o) {
		try {
			if (LocalDate.class.isAssignableFrom(type)) {
				return o.toInstant().atZone(this.getZoneId()).toLocalDate();
			}

			return this.convertTemporalToTemporal(type, o.toInstant().atZone(this.getZoneId()).toLocalDateTime());
		} catch (final DateTimeException e) {
			return null;
		}
	}

	protected <T> TemporalAccessor convertTemporalToTemporal(final Class<T> type, final TemporalAccessor o) {
		if (type.isAssignableFrom(o.getClass())) {
			return o;
		}
		try {
			return this.convertTemporalToTemporal(type, o, type.getDeclaredMethod("from", TemporalAccessor.class));
		} catch (NoSuchMethodException | SecurityException e1) {
			return null;
		}
	}

	private <T> TemporalAccessor convertTemporalToTemporal(final Class<T> type, final TemporalAccessor o, final Method m) {
		try {
			return (TemporalAccessor) m.invoke(null, o);
		} catch (SecurityException | DateTimeException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			return this.convertDateToTemporal(type, this.convertTemporalToCalendar(o).getTime());
		}
	}

	public ZoneId getZoneId() {
		return this.getTimeZone() == null ? ZoneId.systemDefault() : this.getTimeZone().toZoneId();
	}

	@Override
	public Class<?>[] getTypes() {
		return new Class<?>[] {
			Date.class,
			java.sql.Date.class,
			java.sql.Time.class,
			java.sql.Timestamp.class,
			Calendar.class,
			GregorianCalendar.class,
			XMLGregorianCalendar.class,
			TemporalAccessor.class
		};
	}
}
